DATABRICKS_ACCESS_TOKEN = 'DATABRICKS_ACCESS_TOKEN'
DATABRICKS_SERVER_HOSTNAME = 'DATABRICKS_SERVER_HOSTNAME'
DATABRICKS_HTTP_PATH = 'DATABRICKS_HTTP_PATH'

DBX_FORCE_LOCAL = 'DBX_FORCE_LOCAL'

AIRFLOW_UNIQUE_NAME = 'AIRFLOW_UNIQUE_NAME'
